"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from q3dviewer.base_item import BaseItem
from OpenGL.GL import *
from PySide6.QtWidgets import QLabel, QDoubleSpinBox, QLineEdit
from PySide6.QtCore import QRegularExpression
from PySide6.QtGui import QRegularExpressionValidator
import numpy as np


class GridItem(BaseItem):
    def __init__(self, size, spacing, color=np.array([255, 255, 255, 76.5]), offset=np.array([0, 0, 0])):
        super().__init__()
        self.size = size
        self.spacing = spacing
        self.color = color
        self.offset = offset

    def add_setting(self, layout):
        label_size = QLabel("Set size:")
        layout.addWidget(label_size)
        spinbox_size = QDoubleSpinBox()
        spinbox_size.setSingleStep(1.0)
        layout.addWidget(spinbox_size)
        spinbox_size.setValue(self.size)
        spinbox_size.valueChanged.connect(self.set_size)
        spinbox_size.setRange(0, 100000)

        label_spacing = QLabel("Set spacing:")
        layout.addWidget(label_spacing)
        spinbox_spacing = QDoubleSpinBox()
        layout.addWidget(spinbox_spacing)
        spinbox_spacing.setSingleStep(0.1)
        spinbox_spacing.setValue(self.spacing)
        spinbox_spacing.valueChanged.connect(self._on_spacing)
        spinbox_spacing.setRange(0, 1000)

        label_offset = QLabel("Set offset (x;y;z):")
        layout.addWidget(label_offset)
        self.edit_offset = QLineEdit()
        regex = QRegularExpression(r"^-?\d+(\.\d+)?;-?\d+(\.\d+)?;-?\d+(\.\d+)?$")
        validator = QRegularExpressionValidator(regex)
        self.edit_offset.setValidator(validator)
        self.edit_offset.setText(f"{self.offset[0]};{self.offset[1]};{self.offset[2]}")
        self.edit_offset.textChanged.connect(self._on_offset)
        layout.addWidget(self.edit_offset)

    def set_size(self, size):
        self.size = size

    def _on_spacing(self, spacing):
        if spacing > 0:
            self.spacing = spacing

    def _on_offset(self, text):
        try:
            values = list(map(float, text.split(';')))
            if len(values) == 3:
                self.offset = np.array(values)
            else:
                raise ValueError("Offset must have 3 values separated by ';'")
        except ValueError:
            pass

    def set_offset(self, offset):
        if isinstance(offset, np.ndarray) and offset.shape == (3,):
            self.offset = offset
            self.edit_offset.setText(f"{self.offset[0]};{self.offset[1]};{self.offset[2]}")
        else:
            raise ValueError("Offset must be a numpy array with shape (3,)")

    def paint(self):
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glColor4f(self.color[0] / 255.0, self.color[1] / 255.0, self.color[2] / 255.0, self.color[3] / 255.0)
        glBegin(GL_LINES)
        for i in np.arange(-self.size, self.size + self.spacing, self.spacing):
            glVertex3f(i + self.offset[0], -self.size + self.offset[1], self.offset[2])
            glVertex3f(i + self.offset[0], self.size + self.offset[1], self.offset[2])
            glVertex3f(-self.size + self.offset[0], i + self.offset[1], self.offset[2])
            glVertex3f(self.size + self.offset[0], i + self.offset[1], self.offset[2])
        glEnd()
        glDisable(GL_BLEND)





"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from q3dviewer.base_item import BaseItem
from q3dviewer.utils import set_uniform
from OpenGL.GL import *
import numpy as np
from OpenGL.GL import shaders
from PySide6.QtWidgets import QLabel, QDoubleSpinBox


# Vertex and Fragment shader source code
vertex_shader_source = """
#version 330 core
layout(location = 0) in vec3 position;
layout(location = 1) in vec3 color;

out vec3 ourColor;

uniform mat4 view_matrix;
uniform mat4 project_matrix;
uniform mat4 model_matrix;
uniform float size;

void main()
{
    vec3 scaled_position = position * size;
    gl_Position = project_matrix * view_matrix * model_matrix * vec4(scaled_position, 1.0);
    ourColor = color;
}
"""

fragment_shader_source = """
#version 330 core
in vec3 ourColor;
out vec4 color;

void main()
{
    color = vec4(ourColor, 1.0);
}
"""


class AxisItem(BaseItem):
    def __init__(self, size=1.0, width=2):
        super().__init__()
        self.size = size
        self.width = width
        self.model_matrix = np.eye(4, dtype=np.float32)
        self.need_update_setting = True

    def initialize_gl(self):
        # Axis vertices and colors
        self.vertices = np.array([
            # positions         # colors
            [0.0, 0.0, 0.0,    1.0, 0.0, 0.0],  # X axis (red)
            [1.0, 0.0, 0.0,    1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0,    0.0, 1.0, 0.0],  # Y axis (green)
            [0.0, 1.0, 0.0,    0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0,    0.0, 0.0, 1.0],  # Z axis (blue)
            [0.0, 0.0, 1.0,    0.0, 0.0, 1.0],
        ], dtype=np.float32)

        self.vao = glGenVertexArrays(1)
        vbo = glGenBuffers(1)

        glBindVertexArray(self.vao)

        glBindBuffer(GL_ARRAY_BUFFER, vbo)
        glBufferData(GL_ARRAY_BUFFER, self.vertices.nbytes, self.vertices, GL_STATIC_DRAW)

        # Vertex positions
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 24, ctypes.c_void_p(0))
        glEnableVertexAttribArray(0)

        # Vertex colors
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 24, ctypes.c_void_p(12))
        glEnableVertexAttribArray(1)

        # Compile shaders and create shader program
        self.program = shaders.compileProgram(
            shaders.compileShader(vertex_shader_source, GL_VERTEX_SHADER),
            shaders.compileShader(fragment_shader_source, GL_FRAGMENT_SHADER),
        )

        glBindVertexArray(0)

    def add_setting(self, layout):
        label_size = QLabel("Set size:")
        layout.addWidget(label_size)
        spinbox_size = QDoubleSpinBox()
        spinbox_size.setSingleStep(0.1)
        layout.addWidget(spinbox_size)
        spinbox_size.setValue(self.size)
        spinbox_size.valueChanged.connect(self.set_size)
        spinbox_size.setRange(0.0, 100)

        label_width = QLabel("Set width:")
        layout.addWidget(label_width)
        spinbox_width = QDoubleSpinBox()
        layout.addWidget(spinbox_width)
        spinbox_width.setSingleStep(0.1)
        spinbox_width.setValue(self.width)
        spinbox_width.valueChanged.connect(self.set_width)
        spinbox_width.setRange(0, 1000)

    def set_size(self, size):
        self.size = size
        self.need_update_setting = True

    def update_setting(self):
        if not self.need_update_setting:
            return
        glUseProgram(self.program)
        set_uniform(self.program, float(self.size), 'size')
        set_uniform(self.program, self.model_matrix, 'model_matrix')
        glUseProgram(0)
        self.need_update_setting = False

    def set_width(self, width):
        self.width = width
        
    def set_transform(self, transform):
        """
        Set the transformation matrix for the axis item.
        """
        self.model_matrix = transform
        self.need_update_setting = True

    def paint(self):
        self.update_setting()
        glLineWidth(self.width)
        glUseProgram(self.program)
        glBindVertexArray(self.vao)

        view_matrix = self.glwidget().get_view_matrix()
        project_matrix = self.glwidget().get_projection_matrix()
        set_uniform(self.program, view_matrix, 'view_matrix')
        set_uniform(self.program, project_matrix, 'project_matrix')

        glDrawArrays(GL_LINES, 0, 6)

        glBindVertexArray(0)
        glUseProgram(0)
        glLineWidth(1)



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

import numpy as np


def frustum(left, right, bottom, top, near, far):
    # see https://www.khronos.org/registry/OpenGL-Refpages/gl2.1/xhtml/glFrustum.xml
    if near <= 0 or far <= 0 or near >= far or left == right or bottom == top:
        print("Invalid frustum parameters.")
        return None
    matrix = np.zeros((4, 4), dtype=np.float32)
    matrix[0, 0] = 2.0 * near / (right - left)
    matrix[0, 2] = (right + left) / (right - left)
    matrix[1, 1] = 2.0 * near / (top - bottom)
    matrix[1, 2] = (top + bottom) / (top - bottom)
    matrix[2, 2] = -(far + near) / (far - near)
    matrix[2, 3] = -2.0 * far * near / (far - near)
    matrix[3, 2] = -1.0
    return matrix


def rainbow(scalars, scalar_min=0, scalar_max=255):
    range = scalar_max - scalar_min
    values = 1.0 - (scalars - scalar_min) / range
    # values = (scalars - scalar_min) / range  # using inverted color
    colors = np.zeros([scalars.shape[0], 3], dtype=np.float32)
    values = np.clip(values, 0, 1)

    h = values * 5.0 + 1.0
    i = np.floor(h).astype(int)
    f = h - i
    f[np.logical_not(i % 2)] = 1 - f[np.logical_not(i % 2)]
    n = 1 - f

    # idx = i <= 1
    colors[i <= 1, 0] = n[i <= 1] * 255
    colors[i <= 1, 1] = 0
    colors[i <= 1, 2] = 255

    colors[i == 2, 0] = 0
    colors[i == 2, 1] = n[i == 2] * 255
    colors[i == 2, 2] = 255

    colors[i == 3, 0] = 0
    colors[i == 3, 1] = 255
    colors[i == 3, 2] = n[i == 3] * 255

    colors[i == 4, 0] = n[i == 4] * 255
    colors[i == 4, 1] = 255
    colors[i == 4, 2] = 0

    colors[i >= 5, 0] = 255
    colors[i >= 5, 1] = n[i >= 5] * 255
    colors[i >= 5, 2] = 0
    return colors


def euler_to_matrix(rpy):
    roll, pitch, yaw = rpy
    Rx = np.array([[1, 0, 0],
                   [0, np.cos(roll), -np.sin(roll)],
                   [0, np.sin(roll), np.cos(roll)]])
    Ry = np.array([[np.cos(pitch), 0, np.sin(pitch)],
                   [0, 1, 0],
                   [-np.sin(pitch), 0, np.cos(pitch)]])
    Rz = np.array([[np.cos(yaw), -np.sin(yaw), 0],
                   [np.sin(yaw), np.cos(yaw), 0],
                   [0, 0, 1]])
    R = Rz @ Ry @ Rx
    return R


def matrix_to_euler(R):
    sy = np.sqrt(R[0, 0]**2 + R[1, 0]**2)
    singular = sy < 1e-6  # Check for gimbal lock
    if not singular:
        roll = np.arctan2(R[2, 1], R[2, 2])   # X-axis rotation
        pitch = np.arctan2(-R[2, 0], sy)      # Y-axis rotation
        yaw = np.arctan2(R[1, 0], R[0, 0])    # Z-axis rotation
    else:
        # Gimbal lock case
        roll = np.arctan2(-R[1, 2], R[1, 1])
        pitch = np.arctan2(-R[2, 0], sy)
        yaw = 0  # Arbitrarily set yaw to 0

    return np.array([roll, pitch, yaw])


def matrix_to_quaternion(matrix):
    trace = matrix[0, 0] + matrix[1, 1] + matrix[2, 2]
    if trace > 0:
        s = 0.5 / np.sqrt(trace + 1.0)
        w = 0.25 / s
        x = (matrix[2, 1] - matrix[1, 2]) * s
        y = (matrix[0, 2] - matrix[2, 0]) * s
        z = (matrix[1, 0] - matrix[0, 1]) * s
    else:
        if matrix[0, 0] > matrix[1, 1] and matrix[0, 0] > matrix[2, 2]:
            s = 2.0 * np.sqrt(1.0 + matrix[0, 0] - matrix[1, 1] - matrix[2, 2])
            w = (matrix[2, 1] - matrix[1, 2]) / s
            x = 0.25 * s
            y = (matrix[0, 1] + matrix[1, 0]) / s
            z = (matrix[0, 2] + matrix[2, 0]) / s
        elif matrix[1, 1] > matrix[2, 2]:
            s = 2.0 * np.sqrt(1.0 + matrix[1, 1] - matrix[0, 0] - matrix[2, 2])
            w = (matrix[0, 2] - matrix[2, 0]) / s
            x = (matrix[0, 1] + matrix[1, 0]) / s
            y = 0.25 * s
            z = (matrix[1, 2] + matrix[2, 1]) / s
        else:
            s = 2.0 * np.sqrt(1.0 + matrix[2, 2] - matrix[0, 0] - matrix[1, 1])
            w = (matrix[1, 0] - matrix[0, 1]) / s
            x = (matrix[0, 2] + matrix[2, 0]) / s
            y = (matrix[1, 2] + matrix[2, 1]) / s
            z = 0.25 * s
    return np.array([x, y, z, w])


def quaternion_to_matrix(quaternion):
    x, y, z, w = quaternion
    q = np.array(quaternion[:4], dtype=np.float64, copy=True)
    n = np.linalg.norm(q)
    if np.any(n == 0.0):
        raise ZeroDivisionError("bad quaternion input")
    else:
        m = np.empty((3, 3))
        m[0, 0] = 1.0 - 2*(y**2 + z**2)/n
        m[0, 1] = 2*(x*y - z*w)/n
        m[0, 2] = 2*(x*z + y*w)/n
        m[1, 0] = 2*(x*y + z*w)/n
        m[1, 1] = 1.0 - 2*(x**2 + z**2)/n
        m[1, 2] = 2*(y*z - x*w)/n
        m[2, 0] = 2*(x*z - y*w)/n
        m[2, 1] = 2*(y*z + x*w)/n
        m[2, 2] = 1.0 - 2*(x**2 + y**2)/n
        return m


def make_transform(pose, rotation):
    transform = np.eye(4)
    transform[0:3, 0:3] = quaternion_to_matrix(rotation)
    transform[0:3, 3] = pose
    return transform

def makeT(R, t):
    T = np.eye(4)
    T[0:3, 0:3] = R
    T[0:3, 3] = t
    return T

def makeRt(T):
    R = T[0:3, 0:3]
    t = T[0:3, 3]
    return R, t

def hex_to_rgba(hex_color):
    color_flat = int(hex_color[1:], 16)
    red = (color_flat >> 16) & 0xFF
    green = (color_flat >> 8) & 0xFF
    blue = color_flat & 0xFF
    return (red / 255.0, green / 255.0, blue / 255.0, 1.0)

# euler = np.array([1, 0.1, 0.1])
# euler_angles = matrix_to_euler(euler_to_matrix(euler))
# print("Euler Angles:", euler_angles)



from setuptools import setup, find_packages

setup(
    name='q3dviewer',
    version='1.0.7',
    packages=find_packages(),
    include_package_data=True,
    package_data={
        'q3dviewer': ['shaders/*.glsl'],
    },
    install_requires=[
        'numpy',
        'pyside6',
        'PyOpenGL',
        'pillow',
        'meshio',
        'pypcd4',
        'pye57',
        'laspy',
        ],
    entry_points={
        'console_scripts': [
            'cloud_viewer = q3dviewer.tools.cloud_viewer:main',
            'ros_viewer = q3dviewer.tools.ros_viewer:main',
            'mesh_viewer = q3dviewer.tools.mesh_viewer:main',
            'gaussian_viewer = q3dviewer.tools.gaussian_viewer:main',
            'lidar_cam_calib = q3dviewer.tools.lidar_cam_calib:main',
            'lidar_calib = q3dviewer.tools.lidar_calib:main',
        ],
    },
)